@@ -2,9 +2,9 @@
  * Gnome Chemistry Utils
  * object.cc 
  *
- * Copyright (C) 2002-2004
+ * Copyright (C) 2002-2005
  *
- * Developed by Jean Bréfort <jean.brefort@ac-dijon.fr>
+ * Developed by Jean Bréfort <jean.brefort@normalesup.org>
  *
  * This library is free software; you can redistribute it and/or
  * modify it under the terms of the GNU Lesser General Public
@@ -16,12 +16,13 @@
  * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
  * Lesser General Public License for more details.
  *
- * You should have received a copy of the GNU Lesser General Public
- * License along with this library; if not, write to the 
- * Free Software Foundation, Inc., 59 Temple Place - Suite 330,
- * Boston, MA  02111-1307, USA.
+ * You should have received a copy of the GNU Lesser General Public License
+ * along with this program; if not, write to the Free Software
+ * Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301
+ * USA
  */
 
+#include "config.h"
 #include "object.h"
 #include "document.h"
 #include <string>
@@ -35,6 +36,7 @@ Object::Object(TypeId Id)
 	m_Type = Id;
 	m_Id = NULL;
 	m_Parent = NULL;
+	m_Locked = 0;
 }
 
 Object::~Object()
@@ -58,15 +60,17 @@ Object::~Object()
 }
 
 
-void Object::SetId(gchar* Id)
+void Object::SetId (gchar* Id)
 {
-	if (!Id) return;
-	if (m_Id)
-	{
-		if (m_Parent) m_Parent->m_Children.erase(m_Id);
+	if (!Id)
+		return;
+	if (m_Id) {
+		if (!strcmp (Id, m_Id))
+			return;
+		if (m_Parent) m_Parent->m_Children.erase (m_Id);
 		g_free(m_Id);
 	}
-	m_Id = g_strdup(Id);
+	m_Id = g_strdup (Id);
 	if (m_Parent) {
 		Object *parent = m_Parent;
 		m_Parent = NULL;
@@ -127,7 +131,7 @@ void Object::AddChild(Object* object)
 	}
 	else
 	{
-		Object* o = pDoc->GetDescendant(object->m_Id);
+		Object* o = pDoc->RealGetDescendant(object->m_Id);
 		if (o && ((pDoc != object->GetDocument()) || (object != o)))
 		{
 			gchar *buf = pDoc->GetNewId (object->m_Id);
@@ -231,6 +235,7 @@ bool Object::Load (xmlNodePtr node)
 	xmlNodePtr child;
 	Object* pObject;
 
+	m_Locked++;
 	tmp = xmlGetProp (node, (xmlChar*) "id");
 	if (tmp) {
 		SetId ((char*) tmp);
@@ -239,10 +244,16 @@ bool Object::Load (xmlNodePtr node)
 	child = node->children;
 	while (child) {
 		pObject = CreateObject ((const char*) child->name, this);
-		if (!pObject || !pObject->Load (child))
+		if (pObject) {
+			if (!pObject->Load (child))
+				delete pObject;
+		} else {
+			m_Locked--;
 			return false;
+		}
 		child = child->next;
 	}
+	m_Locked--;
 	return true;
 }
 
@@ -252,7 +263,7 @@ bool Object::SaveChildren(xmlDocPtr xml, xmlNodePtr node)
 	xmlNodePtr child;
 	for (i = m_Children.begin(); i != m_Children.end(); i++)
 	{
-		if (child = (*i).second->Save(xml)) xmlAddChild(node, child);
+		if ((child = (*i).second->Save(xml))) xmlAddChild(node, child);
 		else return false;
 	}
 	return true;
@@ -302,9 +313,9 @@ void Object::Transform2D(Matrix2D& m, double x, double y)
 	for (i = m_Children.begin(); i != m_Children.end(); i++) (*i).second->Transform2D(m, x, y);
 }
 
-bool Object::BuildContextualMenu (GtkUIManager *UIManager)
+bool Object::BuildContextualMenu (GtkUIManager *UIManager, Object *object)
 {
-	return (m_Parent)? m_Parent->BuildContextualMenu (UIManager): false;
+	return (m_Parent)? m_Parent->BuildContextualMenu (UIManager, object): false;
 }
 
 void Object::Add(GtkWidget* w)
@@ -376,13 +387,18 @@ TypeId Object::AddType(string TypeName, Object*(*Create)(), TypeId id)
 		NextType = TypeId ((unsigned) NextType + 1);
 	} else
 		typedesc.Id = id;
-	if (TypeNames.capacity() <= id) {
-		size_t max = (((size_t) id / 10) + 1) * 10;
-		TypeNames.reserve (max--);
-		while (max > TypeNames.size())
-			TypeNames.push_back ("");
+	if (TypeNames.capacity() <= typedesc.Id) {
+		size_t max = (((size_t) typedesc.Id / 10) + 1) * 10;
+		TypeNames.resize (max);
 	}
+#if HAS_VECTOR_AT
 	TypeNames.at (typedesc.Id) = TypeName;
+#else
+	vector<string>::iterator it;
+	it = TypeNames.begin ();
+	it += typedesc.Id;
+	TypeNames.insert (it, TypeName);
+#endif
 	return typedesc.Id;
 }
 
@@ -390,7 +406,14 @@ Object* Object::CreateObject(const string& TypeName, Object* parent)
 {
 	TypeDesc& typedesc = Types[TypeName];
 	Object* pObj = (typedesc.Create)? typedesc.Create(): NULL;
-	if (parent && pObj) parent->AddChild(pObj);
+	if (parent && pObj) {
+		if (pObj->m_Id) {
+			char* newId = parent->GetDocument()->GetNewId (pObj->m_Id, false);
+			pObj->SetId (newId);
+			delete [] newId;
+		}
+		parent->AddChild(pObj);
+	}
 	return pObj;
 }
 
@@ -450,6 +473,7 @@ const set<TypeId>& Object::GetRules (TypeId type, RuleId rule)
 
 const set<TypeId>& Object::GetRules (const string& type, RuleId rule)
 {
+	static set<TypeId> noId;
 	TypeDesc& typedesc = Types[type];
 	switch (rule) {
 		case RuleMustContain:
@@ -460,7 +484,10 @@ const set<TypeId>& Object::GetRules (const string& type, RuleId rule)
 			return typedesc.RequiredParents;
 		case RuleMayBeIn:
 			return typedesc.PossibleParents;
+		default:
+			return noId;
 	}
+	return noId;
 }
 
 static void AddAncestorTypes (TypeId type, set<TypeId>& types)
@@ -505,7 +532,7 @@ void Object::EmitSignal (SignalId Signal)
 {
 	Object *obj = NULL;
 	Object *ancestor = this;
-	while (ancestor && ancestor->OnSignal (Signal, obj)) {
+	while (ancestor && !ancestor->IsLocked () && ancestor->OnSignal (Signal, obj)) {
 		obj = ancestor;
 		ancestor = obj->m_Parent;
 	}
@@ -539,3 +566,11 @@ void Object::Unlink (Object *object)
 void Object::OnUnlink (Object *object)
 {
 }
+
+void Object::Lock (bool state)
+{
+	if (state)
+		m_Locked++;
+	else if (m_Locked > 0)
+		m_Locked--;
+}
